extends CharacterBody3D

@export var body : Node
@export var head : Node
@export  var cam : Node
@export var col_shape : Node
@export var col_stand : Node
@export var col_crouch : Node
var tween
#export(NodePath) onready var disable_input_timer = get_node(disable_input_timer) as Timer
#export(NodePath) onready var splash_in_particles = get_node(splash_in_particles) as Particles
#export(NodePath) onready var splash_out_particles = get_node(splash_out_particles) as Particles
#export(NodePath) onready var teleport_particles = get_node(teleport_particles) as Particles


var mouse_sensitivity := 25.0

var stop_speed := 4.0
var move_speed := 6.0
var max_slope_angle := 46.0
var max_stair_slope := 1.0

var gravity := 60.0
var max_fall_speed := 50.0
var jump_height := 2.0

var accel_ground := 10.0
var accel_air := 3.0
var accel_water := 4.0

var friction_ground := 4.0 # was 6.0
var friction_water := 1.0

var step_height := 1

var queue_jump := false
var hold_jump := false
var extra_jumps_left := 0
var extra_jumps_total := 0
var coyote_spent := false
var coyote_wait := 0.0
var coyote_amount := 0.3

var snap_vector := Vector3.ZERO
var grounded := false
var wall := false
var crouch := false
var on_ladder := false

var current_water = null
var swimming := false

var autorun := true
var walk_hold := false
var walk_multiplier := 0.5


const cr_sp = 0.3
const ldr_sp = 0.5

#@onready var head_offset := [head.transform.origin.y,head.transform.origin.z]
@onready var head_start = head.position.y
var head_down = 1.00


var extra_vel := Vector3.ZERO

var input_block := false

func _ready():
	Input.set_mouse_mode(Input.MOUSE_MODE_CAPTURED)

	body.rotation.y = rotation.y
	rotation.y = 0

	jump_reset()


func get_move_direction() -> Vector3:

	var input_dir := Vector2(
		Input.get_axis("move_forward","move_backward"),
		Input.get_axis("move_left", "move_right")
	).limit_length(1.0)

	if !on_ladder and (!swimming or (swimming and grounded and head.rotation.x < 0)):
		return input_dir.x * body.global_transform.basis.z + input_dir.y * body.global_transform.basis.x
	else:
		return input_dir.x * head.global_transform.basis.z + input_dir.y * body.global_transform.basis.x

func _input(event):

	if event is InputEventMouseMotion:
#		if level_ready:
			rotate_look(event.relative * 0.001 * mouse_sensitivity)

	if Input.is_action_just_pressed("crouch"):
#		if grounded:
			if !crouch:
				col_shape.shape.extents = col_crouch.shape.extents
				col_shape.transform.origin = col_crouch.transform.origin
				crouch_tween()
				crouch = true

			else:
				if overlap_check() == []:
					col_shape.shape.extents = col_stand.shape.extents
					col_shape.transform.origin = col_stand.transform.origin
					crouch_tween()
					crouch = false

	if Input.is_action_just_pressed("move_jump") and !crouch:
		queue_jump = true
	if Input.is_action_just_released("move_jump") and !crouch:
		queue_jump = false

	if Input.is_action_just_pressed("autorun"):
		autorun = !autorun
#		Global.notify("Autorun On" if autorun else "Autorun Off")

func _process(delta):
	hold_jump = Input.is_action_pressed("move_jump")
	walk_hold = Input.is_action_pressed("walk")


func crouch_tween():
	if tween != null:
		tween.kill()
	tween = get_tree().create_tween()
	tween.tween_property(
			head,
			"position:y",
#			Vector3(0,head_offset[0],head_offset[1]) if crouch else Vector3(0,-0.75,head_offset[1]),
			head_start - head_down if not crouch else head_start,
			0.2
		)

func rotate_look(amount : Vector2) -> void:
	body.rotation.y -= amount.x
	head.rotation.x = clamp(head.rotation.x - amount.y, -PI * 0.5, PI * 0.5)


func _physics_process(delta):
	grounded = is_on_floor()
	wall = is_on_wall()
#	in_water_check()
	jump_reset()
	extra_vel_func(delta)
	velocity.y = max(-fall_check(), velocity.y - gravity_check() * delta)

	var h_target_dir : Vector3 = get_move_direction() if not input_block else Vector3.ZERO
	if !on_ladder and !swimming:
		if grounded:
			apply_friction(delta)
			accelerate(delta, h_target_dir,
			move_speed * (cr_sp if crouch else 1.0) * walk_check(), #move speed
			accel_ground)
	#		snap_vector = -get_floor_normal()

			snap_vector = -get_floor_normal() - get_platform_velocity() * delta

			# Workaround for sliding down after jump on slope
			if velocity.y < 0:
				velocity.y = 0

			if queue_jump:
#				velocity.y = sqrt(2 * jump_height * abs(gravity))
#				queue_jump = false
#				snap_vector = Vector3.ZERO
				coyote_spent = true
				jump_func()

		else: #if not grounded
			# Workaround for 'vertical bump' when going off platform
			if snap_vector != Vector3.ZERO && velocity.y != 0:
				velocity.y = 0

			accelerate(delta, h_target_dir, move_speed, accel_air) #move speed
			snap_vector = Vector3.ZERO
#			if crouch:
#				if overlap_check() == []:
#					col_stand.disabled = false
#					col_crouch.disabled = true
#					crouch_tween()
#					crouch = false
			coyote_time(delta)
			if queue_jump:
				if velocity.y < 0 and !coyote_spent and coyote_wait < coyote_amount:
					velocity.y = 0
					jump_func()
				elif (coyote_wait > coyote_amount or coyote_spent) and extra_jumps_left > 0:
					velocity.y = 0
					jump_func()
					extra_jumps_left -= 1
					print ("hmm")
#				if 	(velocity.y < 0 and jumps_left == 1 and coyote_wait < coyote_amount) or \
#					(jumps_total > 1 and jumps_left > 0):
	elif on_ladder:
		snap_vector = Vector3.ZERO
		apply_friction(delta)
		accelerate(delta, h_target_dir,
		move_speed * ldr_sp * (0.8 if crouch else 1.0) * walk_check(), #move speed
		accel_ground)
		if test_move(global_transform,Vector3(velocity.x,0,velocity.z)):
			velocity.x  = 0
		if queue_jump:
			var jump_off_dir = body.global_transform.basis.z * move_speed
			if !test_move(global_transform,jump_off_dir):
				velocity += jump_off_dir
			else:
				velocity -= jump_off_dir
			queue_jump = false
	elif swimming:
		snap_vector = Vector3.ZERO
		apply_friction(delta)
		accelerate(delta, h_target_dir,
		move_speed * ldr_sp * (0.7 if crouch else 1.0) * walk_check(), #move speed
		accel_ground)
		if hold_jump:
				velocity.y = sqrt(2 * (jump_height * 0.1) * abs(gravity))
#		if jump_just_released:
#			velocity.y = 0
#		jump_just_released = false

	velocity += extra_vel#, snap_vector, Vector3.UP, false, 4, deg_to_rad(max_slope_angle), false
	move_and_slide()

func accelerate(delta : float, p_target_dir : Vector3, p_target_speed : float, p_accel : float):
	var current_speed : float = velocity.dot(p_target_dir)
	var add_speed : float = p_target_speed - current_speed
	if add_speed > 0:
		var accel_speed : float = min(add_speed, p_accel * delta * p_target_speed)
		velocity += p_target_dir * accel_speed

func apply_friction(delta : float):

	var speed : float = velocity.length()
	if is_zero_approx(speed):
		velocity = Vector3.ZERO
		return

	var drop : float = 0.0
	var control : float = max(speed, stop_speed)

	# ground friction
	if grounded or on_ladder or swimming:
		drop += control * friction_ground * delta

	# water friction not implemented
	var new_speed : float = max(0.0, speed - drop)
	new_speed /= speed

	velocity *= new_speed

func extra_vel_func(delta):
	var speed : float = extra_vel.length()
	if is_zero_approx(speed):
		extra_vel = Vector3.ZERO
		return

	speed = max(0.0, speed - friction_ground)
	extra_vel = extra_vel.normalized() * speed

#func disable_input(time : float = 0.5):
#	input_block = true
#	disable_input_timer.start(time)

func enable_input():
	input_block = false

func overlap_check():
	var space: PhysicsDirectSpaceState3D = get_world_3d().direct_space_state as PhysicsDirectSpaceState3D

	var params: = PhysicsShapeQueryParameters3D.new()
	params.set_shape(col_stand.shape)
	params.set_transform(global_transform)
	params.set_exclude([self])
	params.collision_mask = collision_mask

	var castResult = space.intersect_shape(params)
	return castResult

#func in_water_check():
#	if current_water != null:
#		var vp_size = get_viewport().size
#		var cam_vp = cam.project_ray_origin(vp_size * 0.5)
#		var space: PhysicsDirectSpaceState3D = get_world_3d().direct_space_state as PhysicsDirectSpaceState3D
#		var hit = space.intersect_point ( cam_vp, 1, [self], 4, false, true )
#		if hit != []:
#			if not swimming:
#				splash_in_particles.visible = true
#				splash_in_particles.restart()
#				splash_out_particles.visible = false
#			Global.underwater_screen(true)
#			current_water.draw_order(1)
#			swimming = true
#		else:
#			splash_in_particles.visible = false
#			Global.underwater_screen(false)
#			current_water.draw_order(-1)
#
#func splash_out():
#	splash_out_particles.visible = true
#	splash_out_particles.restart()

func gravity_check():
	if !grounded and !on_ladder and !swimming:
		return gravity
	elif swimming:
		if get_move_direction().length() == 0:
			return 30.0
		else:
			return 0.0
	elif on_ladder:
		return 0.0
	else:
		return gravity

func fall_check():
	if !grounded and !on_ladder and !swimming:
		return max_fall_speed
	elif swimming:
		return max_fall_speed * 0.05
	else:
		return max_fall_speed

func walk_check():
	if (autorun and !walk_hold) or (!autorun and walk_hold):
		return 1.0
	elif (autorun and walk_hold) or (!autorun and !walk_hold):
		return walk_multiplier

func jump_func():
#	if jumps_left > 0:
		velocity.y = sqrt(2 * jump_height * abs(gravity))
		queue_jump = false
		snap_vector = Vector3.ZERO
#		jumps_left -= 1

func coyote_time(delta):
	coyote_wait += 1 * delta
#	if queue_jump and jumps_left > 0 and velocity.y < 0 and coyote_wait < 0.5:
#		velocity.y = 0
#		jump_func()

func jump_reset():
	if grounded:
		coyote_spent = false
		coyote_wait = 0.0
		extra_jumps_left = extra_jumps_total

